import { CosmWasmClient } from "@cosmjs/cosmwasm-stargate";
import { UniversalAddress, encoding, sha256, toChainId, } from "@wormhole-foundation/connect-sdk";
import bs58 from "bs58";
import { CosmwasmAddress } from "./address";
import { CosmwasmPlatform } from "./platform";
import { IBC_TRANSFER_PORT } from "./constants";
export var Gateway;
(function (Gateway) {
    Gateway.name = "Wormchain";
    function gatewayAddress() {
        const contracts = CosmwasmPlatform.contracts.getContracts(Gateway.name);
        return contracts.gateway;
    }
    Gateway.gatewayAddress = gatewayAddress;
    function tokenBridgeAddress() {
        const contracts = CosmwasmPlatform.contracts.getContracts(Gateway.name);
        return contracts.tokenBridge;
    }
    Gateway.tokenBridgeAddress = tokenBridgeAddress;
    function coreAddress() {
        const contracts = CosmwasmPlatform.contracts.getContracts(Gateway.name);
        return contracts.coreBridge;
    }
    Gateway.coreAddress = coreAddress;
    // Returns RPC client for Wormchain
    async function getRpc() {
        const rpcAddress = CosmwasmPlatform.conf[Gateway.name].rpc;
        return await CosmWasmClient.connect(rpcAddress);
    }
    Gateway.getRpc = getRpc;
    async function getTokenBridge(rpc) {
        rpc = rpc ? rpc : await getRpc();
        return CosmwasmPlatform.getTokenBridge(rpc);
    }
    Gateway.getTokenBridge = getTokenBridge;
    async function getIbcBridge(rpc) {
        rpc = rpc ? rpc : await getRpc();
        return CosmwasmPlatform.getIbcBridge(rpc);
    }
    Gateway.getIbcBridge = getIbcBridge;
    // Get the wrapped version of an asset created on wormchain
    // for a given chain
    async function getWrappedAsset(token) {
        const tb = await getTokenBridge();
        const wrappedAsset = await tb.getWrappedAsset(token);
        // Encode the original address to base58 and add it
        // to the factory address for cw20 style factory token address
        const encodedAddress = bs58.encode(wrappedAsset.toUniversalAddress().toUint8Array());
        const factoryAddress = `factory/${gatewayAddress()}/${encodedAddress}`;
        return new CosmwasmAddress(factoryAddress);
    }
    Gateway.getWrappedAsset = getWrappedAsset;
    // Gets the the source channel for outgoing transfers from wormchain
    function getGatewaySourceChannel(chain) {
        const channels = CosmwasmPlatform.getIbcChannels(chain);
        if (!channels)
            throw new Error("No channels configured for chain " + chain);
        if (!(Gateway.name in channels))
            throw new Error("No channel configured for chain " + chain);
        return channels[Gateway.name];
    }
    Gateway.getGatewaySourceChannel = getGatewaySourceChannel;
    // derive the ics20 token denom from the
    // wrapped denom and destination channel
    function deriveIbcDenom(chain, denom) {
        // Otherwise compute the ibc address from the channel and denom
        const channel = getGatewaySourceChannel(chain);
        const hashData = encoding.toUint8Array(`${IBC_TRANSFER_PORT}/${channel}/${denom}`);
        const hash = encoding.hex.encode(sha256(hashData));
        return new CosmwasmAddress(`ibc/${hash.toUpperCase()}`);
    }
    Gateway.deriveIbcDenom = deriveIbcDenom;
    function makeTransferMsg(chain, recipient, fee = 0n, payload, nonce) {
        // Address of recipient is b64 encoded Cosmos bech32 address
        const address = encoding.b64.encode(recipient.toUint8Array());
        const common = {
            chain: toChainId(chain),
            recipient: address,
            fee: fee.toString(),
            nonce: nonce ?? Math.round(Math.random() * 100000),
        };
        const msg = payload
            ? {
                gateway_transfer_with_payload: { ...common, payload: payload },
            }
            : { gateway_transfer: { ...common } };
        return msg;
    }
    Gateway.makeTransferMsg = makeTransferMsg;
    // TODO: make consts
    function parseWormholeMessage(tx) {
        const events = tx.events.filter((ev) => ev.type === "wasm" &&
            ev.attributes[0].key === "_contract_address" &&
            ev.attributes[0].value === coreAddress());
        if (events.length === 0)
            throw new Error("No wormhole message found in tx");
        if (events.length > 1)
            console.error(`Expected single message, found ${events.length}`);
        const [wasm] = events;
        const sequence = wasm.attributes.find((e) => {
            return e.key === "message.sequence";
        }).value;
        const emitter = wasm.attributes.find((e) => {
            return e.key === "message.sender";
        }).value;
        return {
            chain: Gateway.name,
            emitter: new UniversalAddress(emitter),
            sequence: BigInt(sequence),
        };
    }
    Gateway.parseWormholeMessage = parseWormholeMessage;
})(Gateway || (Gateway = {}));
//# sourceMappingURL=gateway.js.map